# -*- coding: utf-8 -*-
from gocept.directory2.domain import Domain, User
import gocept.directory2.admin.domain.testing
import gocept.directory2.admin.testing


class TestDomainCRUD(gocept.directory2.admin.testing.BrowserTestCase,
                     gocept.directory2.admin.domain.testing.DomainMixin):

    def test_domain_list_displays_existing_domains(self):
        self.create_domain()
        browser = self.browser
        browser.open('http://localhost/domain/list')
        self.assertIn('flyingcircus.io', browser.contents)
        self.assertIn('ctheune', browser.contents)

    def test_domain_list_does_not_display_subdomains(self):
        domain, user = self.create_domain()
        Domain.create(
            name='www.flyingcircus.io', uid=user.id, subdomain_of=domain)
        browser = self.browser
        browser.open('http://localhost/domain/list')
        self.assertIn('flyingcircus.io', browser.contents)
        self.assertNotIn('www.flyingcircus.io', browser.contents)

    def test_subdomain_list_displays_only_subdomains_of_current_domain(self):
        self.create_domains_with_subdomains()
        browser = self.browser
        browser.open('http://localhost/domain/list')
        browser.getLink('flyingcircus.io').click()
        browser.getLink('Subdomains', index=0).click()
        self.assertIn('www.flyingcircus.io', browser.contents)
        self.assertNotIn('gocept.de', browser.contents)

    def test_subdomain_list_displays_fancy_booleans(self):
        self.create_domains_with_subdomains()
        Domain.query().filter_by(id=3)[0].sendmail_cw = 0
        browser = self.browser
        browser.open('http://localhost/domain/list')
        browser.getLink('flyingcircus.io').click()
        browser.getLink('Subdomains', index=0).click()
        self.assertEllipsis(u"""...
    <td><span class="bool true">✓</span></td>
    <td><span class="bool false">✗</span></td>...""", browser.contents)

    def test_additional_dns_record_is_displayed_as_pre_text(self):
        self.create_domain()
        Domain.query().filter_by(id=1)[0].freeRecords = "asdf"
        browser = self.browser
        browser.open('http://localhost/domain/1')
        self.assertIn('<pre>asdf</pre>', browser.contents)

    def test_add_domain(self):
        # We need a user which should be selectable during add
        User.create(name='ctheune')
        self.assertEqual(0, Domain.query().count())
        browser = self.browser
        browser.open('http://localhost/domain/list')
        browser.getLink('Add domain').click()
        browser.getControl('Domainname').value = 'flyingcircus.io'
        browser.getControl('User').displayValue = ['ctheune']
        browser.getControl(name='add').click()
        self.assertIn('added successfully', browser.contents)
        self.assertEqual(1, Domain.query().count())
        domain = Domain.query().all()[0]
        self.assertEqual('flyingcircus.io', domain.name)

    def test_add_subdomain(self):
        self.create_domain()
        browser = self.browser
        browser.open('http://localhost/domain/1/list')
        browser.getLink('Add subdomain').click()
        self.assertIn('ctheune', browser.contents)
        with self.assertRaises(LookupError):
            browser.getControl('User')
            browser.getControl('Subdomain')  # Subdomain is added after add
        browser.getControl('Domainname').value = 'www.flyingcircus.io'
        browser.getControl(name='add').click()
        self.assertEqual(
            'http://localhost/domain/1/list', browser.url)
        self.assertIn('www.flyingcircus.io', browser.contents)

    def test_edit_domain(self):
        self.create_domain()
        browser = self.browser
        browser.open('http://localhost/domain/list')
        browser.getLink('flyingcircus.io').click()
        browser.getLink('Edit').click()
        # Displays stored values:
        self.assertIn('flyingcircus.io', browser.contents)
        # Values can be changed:
        browser.getControl('Relative path').value = 'asdf'
        browser.getControl(name='save').click()
        self.assertIn('Updated on', browser.contents)
        # Changed values get persisted:
        browser.getLink('Edit').click()
        self.assertEqual('asdf', browser.getControl('Relative path').value)

    def test_name_and_user_can_not_be_changed_for_domains(self):
        self.create_domain()
        browser = self.browser
        browser.open('http://localhost/domain/1/edit')
        with self.assertRaises(LookupError):
            browser.getControl('Domainname')
            browser.getControl('User')
        self.assertIn('flyingcircus.io', browser.contents)
        self.assertIn('100', browser.contents)  # XXX: show username instead

    def test_delete_domain(self):
        self.create_domain()
        browser = self.browser
        browser.open('http://localhost/domain/list')
        browser.getLink('flyingcircus.io').click()
        browser.getLink('Delete').click()
        self.assertIn(
            'Confirm deletion of domain', browser.contents)
        browser.getControl(name='delete').click()
        self.assertEqual('http://localhost/domain/list', browser.url)
        self.assertNotIn('flyingcircus.io', browser.contents)

    def test_delete_subdomain(self):
        self.create_domains_with_subdomains()
        browser = self.browser
        browser.open('http://localhost/domain/list')
        browser.getLink('flyingcircus.io').click()
        browser.getLink('Subdomains').click()
        browser.getLink('www.flyingcircus.io').click()
        browser.getLink('Delete').click()
        self.assertIn(
            'Confirm deletion of domain', browser.contents)
        browser.getControl(name='delete').click()
        self.assertEqual('http://localhost/domain/1/list', browser.url)
        self.assertNotIn('www.flyingcircus.io', browser.contents)

    def test_domain_detail_page_links_to_user(self):
        self.create_domain()
        browser = self.browser
        browser.open('http://localhost/domain/list')
        browser.getLink('flyingcircus.io').click()
        self.assertEqual('http://localhost/domain/user/100/',
                         browser.getLink('ctheune').url)

    def test_deleting_domain_also_deletes_subdomain(self):
        self.create_domains_with_subdomains()
        browser = self.browser
        browser.open('http://localhost/domain/1/delete')
        self.assertEqual([u'gocept.de', u'www.gocept.de'],
                         [d.name for d in Domain.query().all()])

    def test_deleting_subdomain_does_not_delete_domain(self):
        self.create_domains_with_subdomains()
        browser = self.browser
        browser.open('http://localhost/domain/3/delete')
        self.assertEqual([u'flyingcircus.io', u'gocept.de', u'www.gocept.de'],
                         [d.name for d in Domain.query().all()])
