from ..location import Location
from ..machine import Node, PhysicalMachine, VirtualMachine
from ..resourcegroup import ResourceGroup
from .schema import Relation, ReadOnlyNode
from .view import View, Action
from .widget import RequiredSelectWidget
from colander import SchemaNode, String
from deform.widget import TextAreaWidget
from pyramid.httpexceptions import HTTPFound
from pyramid.view import view_config
import colander
import gocept.directory2.admin.form


@view_config(route_name='machine_list',
             renderer='templates/machine/list.pt')
class ListMachines(View):
    """List known machines."""
    title = 'Machines'
    actions = [
        Action('Add virtual machine', 'icon-plus', 'machine_add_virtual'),
        Action('Add physical machine', 'icon-plus', 'machine_add_physical'),
    ]

    @property
    def machines(self):
        return Node.query()


@view_config(route_name='home',
             renderer='templates/machine/list.pt')
class HomePage(ListMachines):
    pass


@view_config(route_name='machine_view_physical',
             renderer='templates/machine/view_physical.pt')
class ViewPhysical(View):

    @property
    def title(self):
        return self.context.name

    @property
    def actions(self):
        yield Action('Notify all contacts', 'icon-envelope', 'machine_mail',
                     id=self.context.id)
        yield Action('Edit', 'icon-edit',
                     'machine_edit_%s' % self.context.class_,
                     id=self.context.id)
        yield Action('Add NIC', 'icon-plus', 'nic_add', id=self.context.id)
        yield Action('Delete', 'icon-remove', 'machine_confirm_delete',
                     id=self.context.id)


@view_config(route_name='machine_view_virtual',
             renderer='templates/machine/view_virtual.pt')
class ViewVirtual(ViewPhysical):
    pass


class AddSchema(colander.Schema):
    """Schema to add a physical machine."""

    name = SchemaNode(String(), title='Hostname')
    location = SchemaNode(Relation(Location))
    service_description = SchemaNode(
        String(), missing='',
        description='This node provides the following services: '
        '(for customer information)')
    resource_group = SchemaNode(Relation(ResourceGroup))


class AddVirtualSchema(AddSchema):

    kvm_host = SchemaNode(Relation(PhysicalMachine), missing=None)


@view_config(route_name='machine_add_physical')
class AddPhysical(gocept.directory2.admin.form.Add):
    """Add a new physical machine."""

    title = 'Add physical machine'
    schema = AddSchema()
    factory = PhysicalMachine
    next_route = 'machine_list'


@view_config(route_name='machine_add_virtual')
class AddVirtual(gocept.directory2.admin.form.Add):
    """Add a new virtual machine."""

    title = 'Add virtual machine'
    schema = AddVirtualSchema()
    factory = VirtualMachine
    next_route = 'machine_list'


class EditSchema(AddSchema):
    """Edit a physical machine."""

    name = ReadOnlyNode(String(), title='Hostname')


class EditVirtualSchema(EditSchema, AddVirtualSchema):
    pass


@view_config(route_name='machine_edit_physical')
class EditPhysical(gocept.directory2.admin.form.Edit):

    title = 'Edit machine'
    schema = EditSchema()
    next_route = 'machine_view_physical'


@view_config(route_name='machine_edit_virtual')
class EditVirtual(gocept.directory2.admin.form.Edit):

    title = 'Edit machine'
    schema = EditVirtualSchema()
    next_route = 'machine_view_virtual'


@view_config(route_name='machine_confirm_delete')
class ConfirmDelete(gocept.directory2.admin.form.ConfirmDelete):
    """Confirm deletion of physical machine."""

    delete_route = 'machine_delete'

    @property
    def subject(self):
        return 'physical machine "%s"' % self.context.name


@view_config(route_name='machine_delete')
class Delete(gocept.directory2.admin.form.Delete):
    """Actually delete physical machine."""

    next_route = 'machine_list'


class MailSchema(colander.Schema):
    """Schema to add a location."""

    language = SchemaNode(String(),
                          widget=RequiredSelectWidget(
                              values=[('de', 'german'), ('en', 'english')]))
    subject = SchemaNode(String())
    message = SchemaNode(
        String(),
        description='Do not include an address, closing, or footer.',
        widget=TextAreaWidget())


@view_config(route_name='machine_mail')
class MailContacts(gocept.directory2.admin.form.Form):
    """Send a mail to all technical contacts in the machine."""

    buttons = ('send',)
    schema = MailSchema()
    title = 'E-mail all technical contacts'
    next_route = 'machine_list'

    def send_success(self, appstruct):
        self.context.mail_contacts(
            appstruct['language'], appstruct['subject'], appstruct['message'])
        self.request.session.flash('Successfully sent message.', 'success')
        return HTTPFound(location=self.request.route_url(
            self.next_route, id=self.context.id))
