from ..machine import Node
from ..resourcegroup import ResourceGroup
from .schema import ReadOnlyNode
from .view import View, Data, Action
from colander import SchemaNode, String, Int, SequenceSchema
from deform.widget import TextAreaWidget, SelectWidget
from pyramid.view import view_config
import colander
import gocept.directory2.admin.form
import pytz


class ResourceGroupTabs(object):

    @property
    def tabs(self):
        return [
            Action('Info', '', 'resourcegroup_view', id=self.context.id),
            Action('Nodes', '', 'resourcegroup_nodes', id=self.context.id),
            Action('Traffic', '', 'resourcegroup_traffic', id=self.context.id),
            Action('Users', '', 'resourcegroup_users', id=self.context.id)
        ]

    @property
    def actions(self):
        yield Action('Edit', 'icon-edit', 'resourcegroup_edit',
                     id=self.context.id)
        yield Action('Delete', 'icon-remove', 'resourcegroup_confirm_delete',
                     id=self.context.id)


@view_config(route_name='resourcegroup_list',
             renderer='templates/resourcegroup/list.pt')
class ListResourceGroups(View):
    """List known resource groups."""
    title = 'Resource groups'
    actions = [Action('Add resource group', 'icon-plus', 'resourcegroup_add')]

    @property
    def resourcegroups(self):
        for rg in ResourceGroup.query():
            yield Data(title=rg.name,
                       url=self.request.route_url(
                           'resourcegroup_view', id=rg.id))


@view_config(route_name='resourcegroup_view',
             renderer='templates/resourcegroup/view.pt')
class ViewResourceGroup(ResourceGroupTabs, View):
    """View a single resource group."""

    def __call__(self):
        self.title = self.context.name
        self.subtitle = self.context.customer_no
        self.description = self.context.description
        return super(ViewResourceGroup, self).__call__()


def timezone_list():
    """List of all timezones for timezone select widget."""
    all = []
    for country_zones in pytz.country_timezones.values():
        all.extend(country_zones)
    all = [(x, x) for x in all]
    all = sorted(set(all))
    all.insert(0, ('UTC', 'UTC'))
    return all


HourWidget = SelectWidget(values=[(x, x) for x in range(24)])


class AddSchema(colander.Schema):
    """Schema to add a resource group."""

    class TechnicalContacts(SequenceSchema):
        contact = SchemaNode(String())

    name = SchemaNode(String())
    description = SchemaNode(String(), missing='', widget=TextAreaWidget())
    technical_contacts = TechnicalContacts(
        description='Email addresses for technical information.',
        validator=colander.Length(min=1))
    customer_no = SchemaNode(String(), missing='', title='Customer number')
    timezone = SchemaNode(
        String(),
        description='Used for all times regarding this resource group. '
                    'Intended for easier customer communication.',
        default=u'Europe/Berlin',
        widget=SelectWidget(values=timezone_list()))
    maintenance_start = SchemaNode(Int(), title='Start hour for maintenance',
                                   default=22, widget=HourWidget)
    maintenance_end = SchemaNode(Int(), title='End hour for maintenance',
                                 default=5, widget=HourWidget)
    notification_leadtime = SchemaNode(
        Int(), title='Lead time',
        description='Time (in hours) between notifying the customer and the '
                    'beginning of a maintenance activity.',
        default=12,
        validator=colander.Range(0, 120))


@view_config(route_name='resourcegroup_add')
class Add(gocept.directory2.admin.form.Add):

    title = 'Add resource group'
    schema = AddSchema()
    factory = ResourceGroup
    next_route = 'resourcegroup_list'


class EditSchema(AddSchema):

    name = ReadOnlyNode(String())


@view_config(route_name='resourcegroup_edit')
class ResourcegroupEditForm(ResourceGroupTabs,
                            gocept.directory2.admin.form.Edit):

    schema = EditSchema()
    title = 'Edit resource group'
    next_route = 'resourcegroup_view'


@view_config(route_name='resourcegroup_confirm_delete')
class ConfirmDelete(gocept.directory2.admin.form.ConfirmDelete):
    """Confirm deletion of resourcegroup."""

    delete_route = 'resourcegroup_delete'

    @property
    def subject(self):
        return 'resource group "%s"' % self.context.name


@view_config(route_name='resourcegroup_delete')
class Delete(gocept.directory2.admin.form.Delete):
    """Actually delete physical resourcegroup."""

    next_route = 'resourcegroup_list'


@view_config(route_name='resourcegroup_nodes',
             renderer='templates/machine/list.pt')
class ListMachines(ResourceGroupTabs, View):
    """List known machines."""

    @property
    def title(self):
        return self.context.name

    @property
    def machines(self):
        return Node.query().filter_by(resource_group=self.context)
