# -*- coding: utf-8 -*-
# Copyright (c) 2011-2013 gocept gmbh & co. kg
# See also LICENSE.txt

import gocept.directory2.admin.testing
import sqlalchemy.exc


class UserTests(gocept.directory2.admin.testing.BrowserTestCase):

    def setUp(self):
        super(UserTests, self).setUp()
        self.mail_generator = (
            chr(i) + '@example.org' for i in xrange(97, 123))

    def create_user(self, uid, **kw):
        from gocept.directory2.user import HumanUser
        factory = kw.get('factory', HumanUser)
        user = factory.create(uid=uid, firstname='Hans', lastname='Wurst')
        if isinstance(user, HumanUser):
            user.mail = kw.get('mail') or self.mail_generator.next()
        for key, value in kw.iteritems():
            setattr(user, key, value)
        return user

    def test_human_users_can_be_created_in_management_ui(self):
        from gocept.directory2.user import User
        self.assertEqual([], User.query().all())
        b = self.browser
        b.open('http://localhost/')
        b.getLink('Users').click()
        b.getLink('Add human user').click()
        self.assertFalse('password' in b.contents.lower())
        b.getControl('Login name').value = 'hans'
        b.getControl('E-mail address').value = 'hans@example.com'
        self.assertTrue('Login shell' in b.contents)
        self.assertFalse('"form.resource_groups"' in b.contents)
        b.getControl('Firstname').value = 'Hans'
        b.getControl('Lastname').value = 'Wurst'
        b.getControl(name="add").click()
        self.assertEqual(1, len(User.query().all()))
        user = User.query().first()
        self.assertEqual('hans', user.uid)
        self.assertEqual('hans@example.com', user.mail)
        self.assertEqual('/home/hans', user.home_directory)
        self.assertTrue(user.password is None)

    def test_adding_user_with_existing_login_raises_validation_error(self):
        self.create_user('hans')
        b = self.browser
        b.open('http://localhost/user/add_human')
        b.getControl('Login name').value = 'hans'
        b.getControl('E-mail address').value = 'hans@example.com'
        b.getControl('Firstname').value = 'Hans'
        b.getControl('Lastname').value = 'Wurst'
        with self.assertRaisesRegexp(
                sqlalchemy.exc.IntegrityError,
                'Key \(uid\)=\(hans\) already exists.'):
            b.getControl(name='add').click()

    def test_user_is_displayed_in_management_ui(self):
        self.create_user(
            'hans', mail='hans@example.org',
            firstname=u'Hans', lastname=u'Müller', login_shell='/bin/sh')
        b = self.browser
        b.open('http://localhost/')
        b.getLink('Users').click()
        self.assertEllipsis(u"""...
     <tr>
       <td><a href="http://localhost/user/human/1">hans</a></td>
       <td>hans@example.org</td>
       <td>Hans M\xfcller</td>
       <td>human</td>
     </tr>...""", b.contents)
        b.getLink('hans').click()
        self.assertEllipsis(u"""...
     <dl>
       <dt>Numerical UID</dt>
       <dd>1000</dd>
       <dt>E-mail address</dt>
       <dd>hans@example.org</dd>
       <dt>Login shell</dt>
       <dd>/bin/sh</dd>
       <dt>Home directory</dt>
       <dd>/home/hans</dd>
       <dt>Password</dt>
       <dd>-</dd>
       <dt>Super-Admin</dt>
       <dd>-</dd>
       <dt>SSH public keys</dt>
       <dd>0</dd>
     </dl>...""", b.contents.decode('utf-8'))

    def test_human_user_can_be_edited_in_management_ui_except_uid(self):
        user = self.create_user('hans', mail='hans@example.org')
        b = self.browser
        b.open('http://localhost/user/human/1/edit')
        self.assertRaises(LookupError, b.getControl, 'Login name')
        self.assertEqual(
            'hans@example.org', b.getControl('E-mail address').value)
        b.getControl('E-mail address').value = 'klaus@example.org'
        b.getControl('Firstname').value = 'Klaus'
        b.getControl('Lastname').value = 'Meier'
        b.getControl('Password').value = 'asdf'
        b.getControl(name="save").click()
        self.assertEllipsis('...klaus@example.org...', b.contents)
        from gocept.directory2.user import User
        users = list(User.query().all())
        self.assertEqual(1, len(users))
        user = users[0]
        self.assertEqual('hans', user.uid)
        self.assertEqual('klaus@example.org', user.mail)
        self.assertTrue(user.password.startswith('{CRYPT}$6$'))
        self.assertTrue(user.check_password('asdf'))

    def test_leaving_password_field_unchanged_does_not_change_password(self):
        from gocept.directory2.user import User
        user = self.create_user('hans', mail='hans@example.org')
        user.password = 'asdf'
        b = self.browser
        b.open('http://localhost/user/human/1/edit')
        b.getControl(name="save").click()
        user = User.query().first()
        self.assertNotEllipsis('...<ul class="errors">...', b.contents)
        self.assertTrue(user.check_password('asdf'))

    def test_user_can_be_deleted_in_management_ui(self):
        from gocept.directory2.user import User
        self.create_user('hans')
        self.create_user('klaus')
        b = self.browser
        b.open('http://localhost/')
        b.getLink('Users').click()
        b.getLink('klaus').click()
        b.getLink('Delete').click()
        self.assertIn(
            'Confirm deletion of User "Hans Wurst"', b.contents)
        b.getControl(name='delete').click()
        self.assertEqual('http://localhost/user/list', b.url)
        self.assertEqual(
            ['hans'], list(u.uid for u in User.query().all()))
