from . import utils
import gocept.directory2.machine
from .db import Object
from sqlalchemy import Boolean, Column, Integer, DateTime, Interval
from sqlalchemy import ForeignKey, String, Text
from sqlalchemy.orm import relation
import datetime
import jinja2
import pytz
import random
import re
import zope.component
import zope.interface


class Activity(Object):

    __tablename__ = 'maintenance_activities'

    starts = Column(DateTime)
    duration = Column(Interval)
    estimated = Column(Interval)

    comment = Column(Text)
    reference_id = Column(String)

    ended = Column(Boolean)
    result = Column(String)  # XXX enum

    window_id = Column(Integer, ForeignKey('maintenance_windows.id'))
    window = relation('Window', backref='activities')

    node_id = Column(Integer, ForeignKey('nodes.id'), nullable=False)
    node = relation('Node', backref='activities')

    @property
    def ends(self):
        return self.starts + self.estimated

    @property
    def starts_local(self):
        if self.starts is None:
            return None
        tz = pytz.timezone(self.node.resource_group.timezone)
        return tz.normalize(self.starts.astimezone(tz))

    def end(self, result, duration=None):
        """Mark this activity as ended."""
        self.ended = True
        if duration is None:
            duration = utils.now() - self.starts
        self.duration = duration
        self.result = result
        self.window.reschedule()

    def unschedule(self):
        if self.ended:
            raise RuntimeError(
                'Activity %s has already ended and can not be unscheduled')
        window = self.window
        window.activities.remove(self)
        window.reschedule()
        notifier = zope.component.getUtility(INotifier)
        notifier.queue(window)
        self.window = None
        self.starts = None
        self.ended = None
        self.result = None
        self.duration = None

    def overlaps(self, other):
        return utils.ranges_overlap(
            (self.starts, self.ends),
            (other.starts, other.ends))


class Window(Object):

    __tablename__ = 'maintenance_windows'

    announced = Column(Boolean)
    comment = Column(Text)
    starts = Column(DateTime)
    ends = Column(DateTime)

    node_id = Column(Integer, ForeignKey('nodes.id'), nullable=False)
    node = relation('Node', backref='windows')

    @property
    def sorted_activities(self):
        return sorted(self.activities, key=lambda a: a.starts)

    @property
    def starts_local(self):
        if self.starts is None:
            return None
        tz = pytz.timezone(self.node.resource_group.timezone)
        return tz.normalize(self.starts.astimezone(tz))

    @property
    def ends_local(self):
        if self.ends is None:
            return None
        tz = pytz.timezone(self.node.resource_group.timezone)
        return tz.normalize(self.ends.astimezone(tz))

    @property
    def _scheduling_allowed(self):
        notification_limit = utils.now() + datetime.timedelta(
            seconds=self.node.resource_group.notification_leadtime * 60 * 60)
        if self.ended:
            # We never schedule into windows that are already ended.
            return False
        if self.starts > notification_limit:
            # We can always use a window that can still be announced
            return True
        if self.announced:
            # We can also use a window that has been announced.
            return True
        # Any other cases, we don't consider them valid anymore. This can
        # be optimized to reduce latency in some cases, but we'll see
        # whether this is really necessary.
        return False

    def schedule(self, activity):
        """Schedule an activity into this window."""
        if not self._scheduling_allowed:
            return

        if not self.activities:
            # Randomize the first possible time within the first 30 minutes of
            # the window or the remaining time if the activity is longer than
            # window duration - 30 minutes.
            activity.starts = self.starts
            window_duration = (self.ends - self.starts).seconds
            random_limit = window_duration - activity.estimated.seconds
            random_limit = max(random_limit / 60, 0)
            random_limit = min(30, random_limit)
            activity.starts += datetime.timedelta(
                seconds=random.randint(0, random_limit) * 60)
        else:
            # Try appending directly to the end of the last scheduled activity
            activities = list(self.sorted_activities)
            activity.starts = activities[-1].ends

        now = utils.now()
        if activity.starts < now:
            activity.starts = now

        if activity.ends > self.ends:
            # Do not allow activities to knowingly run longer than the window.
            activity.starts = None
            return

        # Accept
        self._materialize(self.node)
        self.activities.append(activity)
        activity.window = self

        notifier = zope.component.getUtility(INotifier)
        notifier.queue(self)

    def _materialize(self, node):
        if self.node is None:
            self.node = node
        else:
            assert self.node is node

    def cancel(self):
        for activity in list(self.activities):
            activity.unschedule()
        self.node.windows.remove(self)

    @property
    def can_be_cancelled(self):
        for activity in self.activities:
            if activity.ended:
                return False
        return True

    @property
    def ended(self):
        # The window remains in "not ended" state until all activites are done.
        for activity in self.activities:
            if not activity.ended:
                return False
        # And if all activities are done we close the window if it's end time
        # has passed.
        return self.ends < utils.now()

    def overlaps(self, other):
        return utils.ranges_overlap(
            (self.starts, self.ends),
            (other.starts, other.ends))

    def reschedule(self):
        # Try to reschedule activities and remove gaps between 'ended' and the
        # next 'starts'.
        activities = list(self.sorted_activities)
        for activity, previous in zip(activities, [None] + activities):
            if activity.ended:
                # Don't reschedule activities that are already ended
                continue
            if previous is None:
                # Can't reschedule an activity if it's the first
                continue
            if previous.ended:
                activity.starts = previous.starts + previous.duration
            else:
                activity.starts = previous.ends
            activity.starts = max(activity.starts, utils.now())


class Planning(object):

    @property
    def pending_activities(self):
        return filter(
            lambda a: a.window is None and not a.ended, self.activities)

    @property
    def unfinished_windows(self):
        return [window for window in self.windows
                if not window.ended]

    @property
    def ended_activities(self):
        return filter(lambda a: a.ended, self.activities)

    def activity_by_reference_id(self, reference_id):
        for activity in self.activities:
            if activity.reference_id == reference_id:
                return activity
        raise KeyError(reference_id)

    def suggest_windows(self, ignore_auto_schedule=False):
        """Enumerate maintenance windows."""
        # TIMEZONE ISSUE: This method has a known issue that is hard to counter
        # and we have decide to only narrow the vector of attack:
        # If the search for a window due to overlapping into the future happens
        # to cross a DST switch then the windows generated for the time after
        # the switch will be off by one hour.
        # We have remedied this a little bit by preferring to use 'now'
        # instead of catching up with the last window, if the last window was
        # in the past.

        now = utils.now()

        totallimit = now + datetime.timedelta(days=30)
        # Suggest already materialized windows first.
        last_window = pytz.UTC.localize(datetime.datetime.min)
        for window in sorted(self.windows, key=lambda w: w.starts):
            yield window
            last_window = window.starts

        # Do not create windows that are earlier than the last window already
        # materialized. This allows us to create a window that eats up all
        # requests until a specific time in the future.
        now = max(last_window, now)

        if not ignore_auto_schedule and not self.auto_schedule:
            return

        # Keep suggesting future windows as needed. We need to use
        # a local time here as the resource group stores the start/end times
        # in local time. However, before letting it out of this scope, we
        # ensure to convert it back to UTC.
        tz = pytz.timezone(self.resource_group.timezone)
        start = now.astimezone(tz)
        start = start.replace(
            hour=self.resource_group.maintenance_start,
            minute=0, second=0, microsecond=0)

        if start <= now:
            # Replace caused us to slip into the past on the current day.
            start += datetime.timedelta(days=1)
        while start < totallimit:
            end = start + datetime.timedelta(
                seconds=self.resource_group.maintenance_duration * 60 * 60)
            window = Window()
            window.starts = start.astimezone(pytz.UTC)
            window.ends = end.astimezone(pytz.UTC)
            window.node = self
            if window._scheduling_allowed:
                yield window
            start += datetime.timedelta(days=1)

    def register_window(self, window):
        """Schedule a new window for a specific time period.

        Automatically schedules any pending activities.

        """
        for rival in self.windows:
            if rival.overlaps(window):
                raise ValueError(rival)
        window._materialize(self)
        self.schedule_pending_activities(window)
        return window

    def schedule_pending_activities(self, window):
        for activity in self.pending_activities:
            window.schedule(activity)

    def schedule_activity(self, activity):
        assert activity in self.activities
        windows = self.suggest_windows()
        while activity.starts is None:
            try:
                windows.next().schedule(activity)
            except StopIteration:
                break

    def register_activity(self, request):
        """Register or update a maintenance activity request."""
        try:
            # If this request was already recorded/scheduled, then
            # update the existing activity
            activity = self.activity_by_reference_id(request.reference_id)
            if request.estimated > activity.estimated:
                activity.unschedule()
            activity.comment = request.comment
            activity.estimated = request.estimated
        except KeyError:
            # If the request was not recorded previously, then record it as an
            # activity now.
            self.activities.append(request)
            activity = request
        self.schedule_activity(activity)
        return activity


class INotifier(zope.interface.Interface):
    pass


class Notifier(object):

    zope.interface.implements(INotifier)

    def __init__(self):
        self._queue = []

    def queue(self, window):
        self._queue.append(window)

    def render_message(self, window):
        """Construct maintenance notification and return (subject, body)."""
        jinja_env = jinja2.Environment(
            loader=jinja2.PackageLoader('gocept.directory2',
                                        'maintenance_templates'),
            trim_blocks=True, undefined=jinja2.StrictUndefined,
            line_comment_prefix='###')
        activities = window.sorted_activities
        data = dict(
            starts=window.starts_local.strftime('%Y-%m-%d %H:%M'),
            ends=window.ends_local.strftime('%Y-%m-%d %H:%M'),
            timezone=window.node.resource_group.timezone,
            node_name=window.node.name,
            redundancy_group=window.node.redundancy_group,
            service_description=window.node.service_description,
            comment=window.comment,
            activities=[dict(
                starts=a.starts_local.strftime('%H:%M'),
                comment=a.comment,
                estimated=utils.format_interval(a.estimated))
                for a in activities])
        if activities:
            data['first_start'] = activities[0].starts_local.strftime(
                '%Y-%m-%d %H:%M')
            data['last_end'] = (
                activities[-1].starts_local +
                activities[-1].estimated).strftime('%Y-%m-%d %H:%M')
        tmpl = jinja_env.get_template('announcement')
        subject = 'Maintenance scheduled for %(starts)s %(timezone)s' % data
        return subject, tmpl.render(data)

    def process(self, window):
        subject, body = self.render_message(window)
        window.node.mail_contacts('en', subject, body)
        window.announced = True


# XXX
def schedule_added_window(window, event):
    notifier = None  # XXX
    notifier.queue(window)


class ScheduleEntry(object):

    start = None

    def __init__(self, node, start):
        self.node = node
        self.start = start

    def apply(self, comment):
        for window in self.node.suggest_windows(ignore_auto_schedule=True):
            if window.starts >= self.start:
                window.comment = comment
                self.node.register_window(window)
                break


class FixedScheduleEntry(object):

    end = None

    def __init__(self, node, start, end):
        self.node = node
        self.start = start
        self.end = end

    def apply(self, comment):
        window = Window(self.start, self.end)
        window.comment = comment
        self.node.register_window(window)


class ScheduleStrategy(object):

    first_day = None

    def __init__(self):
        self.nodes = []

    def get_date(self, day_of_week, hour):
        day = self.first_day + datetime.timedelta(days=day_of_week)
        return day.replace(hour=hour)

    def select(self, node):
        raise NotImplemented()

    def plan(self):
        raise NotImplemented()


class EarlyInfrastructure(ScheduleStrategy):
    """This strategy updates all infrastructure machines in a window between
    10am and 4pm.
    """

    automation_classes = set([
        'role::puppetmaster',
        'role::nagios',
        'role::portage_binhost'])

    def select(self, node):
        return node.puppet_classes.intersection(self.automation_classes)

    def plan(self):
        start = self.get_date(0, 9)
        end = start + datetime.timedelta(seconds=6 * (60 * 60))
        for node in self.nodes:
            yield FixedScheduleEntry(node, start, end)


class InternalTestMachines(ScheduleStrategy):
    """Internal test machines are updated on monday evening."""

    is_test_machine = re.compile('^(test|lit(dev|stag|prod))[0-9]{2}$')

    def select(self, node):
        return self.is_test_machine.match(node.name)

    def plan(self):
        start = self.get_date(0, 17)
        end = start + datetime.timedelta(seconds=6 * (60 * 60))
        for node in self.nodes:
            yield FixedScheduleEntry(node, start, end)


class CustomerTestMachines(ScheduleStrategy):
    """Those machines are used to provide a minimal set of 'real' machines for
    Monday evening, the others for tuesday."""

    is_test_machine = re.compile('^.+(test|dev|stag|staging)[0-9]{2}$')

    def select(self, node):
        return self.is_test_machine.match(node.name)

    def choose_representative_nodes(self):
        self.representative = []
        self.others = list(self.nodes)

        required_classes = set()
        for node in self.nodes:
            required_classes.update(node.puppet_classes)

        while required_classes:
            max_node, max_count = None, 0
            for node in self.others:
                this_count = len(
                    node.puppet_classes.intersection(required_classes))
                if this_count > max_count:
                    max_node = node
                    max_count = this_count
            self.representative.append(max_node)
            self.others.remove(max_node)
            required_classes.difference_update(max_node.puppet_classes)

    def plan(self):
        self.choose_representative_nodes()

        monday_evening = self.get_date(0, 17)
        for node in self.representative:
            yield ScheduleEntry(node, monday_evening)

        tuesday_evening = self.get_date(1, 17)
        for node in self.others:
            yield ScheduleEntry(node, tuesday_evening)


class ProductionVMs(ScheduleStrategy):
    """ Provide a reasonable 'few, more, many' strategy for customer VMs.

    Only select VMs that don't have test systems early, honour the additional
    effort for VMs that have test systems and move them to a later date.

    Try to keep them together 'by resource group'

    """

    def select(self, node):
        return isinstance(node, gocept.directory2.machine.VirtualMachine)

    def plan(self):
        # Start with everything in the 'many' group, then redistribute.
        self.few = []
        self.more = []
        self.many = list(self.nodes)

        limit_few = int(0.05 * len(self.nodes))
        limit_more = int(0.3 * len(self.nodes))

        directory = self.nodes[0].location.directory
        resource_group_names = [rg.name for rg in directory.resource_groups]

        # Create a list of all tested resource groups
        test_group = re.compile('^(.+)(test|dev|stag|staging)$')
        tested_resource_groups = set()
        for name in resource_group_names:
            if not test_group.match(name):
                continue
            production_name = test_group.sub(name, '\1')
            if production_name in resource_group_names:
                tested_resource_groups.add(production_name)

        # Move a number of untested machines to the 'few' group
        for node in list(self.many):
            if node.resource_group.name not in tested_resource_groups:
                self.few.append(node)
                self.many.remove(node)
            if len(self.few) >= limit_few:
                break
        # Move a number of machines to the 'more' group
        for node in list(self.many):
            self.more.append(node)
            self.many.remove(node)
            if len(self.more) >= limit_more:
                break

        tuesday_evening = self.get_date(1, 17)
        for node in self.few:
            yield ScheduleEntry(node, tuesday_evening)

        wednesday_evening = self.get_date(2, 17)
        for node in self.more:
            yield ScheduleEntry(node, wednesday_evening)

        thursday_evening = self.get_date(3, 17)
        for node in self.many:
            yield ScheduleEntry(node, thursday_evening)


class KVMHosts(ScheduleStrategy):
    """KVM hosts all on saturday in a limited window."""

    def select(self, node):
        return 'role::kvm_host' in node.puppet_classes

    def plan(self):
        saturday_morning = self.get_date(5, 9)
        for node in self.nodes:
            yield FixedScheduleEntry(
                node, saturday_morning,
                saturday_morning + datetime.timedelta(seconds=4 * 60 * 60))


class Routers(ScheduleStrategy):
    """One router on monday, the others on Saturday."""

    def select(self, node):
        return 'role::router' in node.puppet_classes

    def plan(self):
        first = self.nodes.pop()
        yield ScheduleEntry(first, self.get_date(0, 13))
        saturday_afternoon = self.get_date(5, 13)
        for node in self.nodes:
            yield FixedScheduleEntry(
                node, saturday_afternoon,
                saturday_afternoon + datetime.timedelta(seconds=4 * 60 * 60))


class Storages(ScheduleStrategy):
    """One storage on Monday, the others distributed over
    Tuesday, Wednesday, and Thursday.
    """

    def select(self, node):
        return 'role::storageserver' in node.puppet_classes

    def plan(self):
        yield ScheduleEntry(self.nodes.pop(), self.get_date(0, 13))

        limit_tuesday = max(1, int(0.2 * len(self.nodes)))
        limit_wednesday = max(1, int(0.3 * len(self.nodes)))

        for i in range(min(len(self.nodes), limit_tuesday)):
            yield ScheduleEntry(self.nodes.pop(), self.get_date(1, 17))

        for i in range(min(len(self.nodes), limit_wednesday)):
            yield ScheduleEntry(self.nodes.pop(), self.get_date(2, 17))

        for node in self.nodes:
            yield ScheduleEntry(node, self.get_date(3, 17))


class GenericDevices(ScheduleStrategy):

    def select(self, node):
        return (not node.puppet_classes or
                set(['role::generic']) == set(node.puppet_classes))

    def plan(self):
        for node in self.nodes:
            yield ScheduleEntry(node, self.get_date(5, 17))


class Restricted(ScheduleStrategy):

    def select(self, node):
        return ('role::restricted' in node.puppet_classes)

    def plan(self):
        for node in self.nodes:
            yield ScheduleEntry(node, self.get_date(3, 17))


class CustomerPhysical(ScheduleStrategy):

    def select(self, node):
        return (not isinstance(node, gocept.directory2.machine.VirtualMachine)
                and node.directory_ring >= 1)

    def plan(self):
        for node in self.nodes:
            yield ScheduleEntry(node, self.get_date(3, 17))


class MaintenanceWeekScheduler(object):
    """Schedule a maintenance week.

    Various strategies can suggest windows (fixed or according to the VMs
    automatic schedule) for each node.

    If a node gets assigned multiple windows then only the earliest window is
    chosen.

    """

    STRATEGIES = [ProductionVMs, KVMHosts, Routers, CustomerPhysical,
                  InternalTestMachines, CustomerTestMachines,
                  EarlyInfrastructure, Storages, GenericDevices, Restricted]

    def __init__(self, location):
        self.location = location
        self.nodes = list(self.location.nodes)
        self.entries = []
        # Shuffle nodes so we don't get accidental bias when selecting sample
        # sub-sets.
        self.nodes = random.shuffle(self.nodes)
        self.strategies = []
        for strategy in self.STRATEGIES:
            self.strategies.append(strategy())

    def schedule(self, monday, note):
        assert monday.isoweekday() == 1
        assert not hasattr(monday, 'date')

        monday = datetime.datetime(monday.year, monday.month, monday.day)
        monday = pytz.UTC.localize(monday)
        self.monday = monday

        # Allow each strategy to select nodes
        for strategy in self.strategies:
            strategy.first_day = self.monday
            strategy.nodes = [
                node for node in self.location.nodes
                if strategy.select(node)]

        # Ask each strategy for a schedule, prefer earlier entries over later
        # entries.
        for strategy in self.strategies:
            self.entries.extend(strategy.plan())

        # Sort schedule by start date, eliminate later duplicates.
        self.entries.sort(key=lambda x: x.start)

        # Check that all nodes have been scheduled.
        unscheduled_nodes = (set(self.location.nodes) -
                             set(x.node for x in self.entries))
        if unscheduled_nodes:
            raise RuntimeError(
                'Unscheduled Nodes', [x.name for x in unscheduled_nodes])

        # Apply all schedule entries.
        scheduled_nodes = set()
        for entry in self.entries:
            if entry.node in scheduled_nodes:
                continue
            entry.apply(note)
            scheduled_nodes.add(entry.node)
