from sqlalchemy import Column, String, Integer
import gocept.directory2.db


class NumberRange(gocept.directory2.db.Object):
    """Abstract base class.

    In a sub-class you have to set the `type`, `min`, `max` attributes and
    implement the `is_available` method.

    """

    type = NotImplemented  # name of range
    min = NotImplemented  # start of range
    max = NotImplemented  # end of range (included)

    __tablename__ = 'numberranges'
    type_ = Column(String, unique=True)
    next_number = Column(Integer)

    def is_available(self, number):
        """Tells whether a number is currently not used."""
        # To be implemented in concrete sub-class.
        raise NotImplementedError()

    @classmethod
    def create(cls):
        numberrange = super(NumberRange, cls).create()
        numberrange.type_ = cls.type
        numberrange.next_number = cls.min
        return numberrange

    @property
    def stop(self):
        return self.max + 1

    @property
    def start(self):
        return self.min

    def next(self):
        """Produces numerical IDs for a class of objects.

        Guarantees that no id will be dealt out more than once within the same
        class of objects (both with respect to ids that have been assigned
        earlier and simultaneously generated new ids). Yields a different
        number from each call, regardless of whether it will actually be
        assigned. Uses the whole available number space between the minimum
        and maximum but tries to avoid yielding numbers that were freed
        recently.

        Uses state stored as in RDB.

        """
        def inc(candidate):
            candidate += 1
            if candidate == self.stop:
                candidate = self.start
            return candidate

        candidate = self.next_number
        while True:
            if self.is_available(candidate):
                self.next_number = inc(candidate)
                return candidate
            candidate = inc(candidate)
            if candidate == self.next_number:
                raise RuntimeError('%r space exhausted.' % self.type)


def NumberRangeFactory(cls):
    """Returns a Factory to get the singleton of a specific number range."""
    def factory():
        query = cls.query().filter_by(type_=cls.type)
        if query.count():
            return query.first()
        return cls.create()
    return factory
