from ..location import Location
from ..machine import VirtualMachine, Node
from ..maintenance import Window, Activity
from ..resourcegroup import ResourceGroup
import datetime
import gocept.directory2.machine
import gocept.directory2.maintenance
import gocept.directory2.testing
import mock
import pytz
import random
import unittest


class MaintenanceTests(unittest.TestCase):

    layer = gocept.directory2.testing.BASE_APPLICATION_LAYER

    def setUp(self):
        self.rg = ResourceGroup.create(name='sample')
        self.rg.timezone = 'UTC'
        self.location = Location.create(name='whq')
        self.node = Node.create(
            name='test00',
            location=self.location,
            resource_group=self.rg,
            auto_schedule=True,
            service_description='Testing purposes')
        self.now_patcher = mock.patch('gocept.directory2.utils.now')
        self.now = self.now_patcher.start()
        self.now.return_value = datetime.datetime(
            2011, 7, 1, 19, 1, 2, 3, pytz.UTC)
        random.seed(12)

    def tearDown(self):
        self.now_patcher.stop()

    def assertInRange(self, expected, got):
        self.assertTrue(
            expected[0] <= got <= expected[1],
            "%s not in range (%s, %s)" % (got, expected[0], expected[1]))

    def test_no_suggestion_if_automatic_scheduling_disabled(self):
        self.node.auto_schedule = False
        request = Activity()
        request.duration = datetime.timedelta(seconds=5)
        request.comment = 'test'
        request.reference_id = 'A'
        suggestion = self.node.register_activity(request)
        self.assertEquals(None, suggestion.starts)
        self.assertEquals(None, suggestion.window)

    def test_register_activity_no_duplicates(self):
        self.now.return_value = datetime.datetime(
            2011, 7, 8, 10, 0, 0, 0, pytz.UTC)
        self.node.resource_group.notification_leadtime = 4
        self.node.resource_group.maintenance_start = 21
        self.node.resource_group.maintenance_end = 23
        estimated_duration = datetime.timedelta(seconds=30 * 60)

        expected_range = [
            datetime.datetime(2011, 7, 8, 21, 0, 0, 0, pytz.UTC),
            datetime.datetime(2011, 7, 8, 22, 30, 0, 0, pytz.UTC)]

        request1 = Activity()
        request1.estimated = estimated_duration
        request1.reference_id = 'A'
        suggestion1 = self.node.register_activity(request1)
        request2 = Activity()
        request2.estimated = estimated_duration
        request2.reference_id = 'B'
        suggestion2 = self.node.register_activity(request2)
        self.assertNotEquals(suggestion1.starts, suggestion2.starts)
        self.assertInRange(expected_range, suggestion1.starts)
        self.assertInRange(expected_range, suggestion2.starts)

    def test_register_activity_first_is_random(self):
        self.now.return_value = datetime.datetime(
            2011, 7, 8, 10, 0, 0, 0, pytz.UTC)
        self.node.resource_group.notification_leadtime = 4
        self.node.resource_group.maintenance_start = 21
        self.node.resource_group.maintenance_end = 23
        estimated_duration = datetime.timedelta(seconds=30 * 60)

        expected_range = [
            datetime.datetime(2011, 7, 8, 21, 0, 0, 0, pytz.UTC),
            datetime.datetime(2011, 7, 8, 21, 30, 0, 0, pytz.UTC)]

        start_times = []
        for request in range(4):
            r = Activity()
            r.estimated = estimated_duration
            s = self.node.register_activity(r)
            start_times.append(s.starts)
            s.unschedule()
        self.assertTrue(len(set(start_times)) >= 2)
        for time in start_times:
            self.assertInRange(expected_range, time)

    def test_unschedule_activity(self):
        r = Activity()
        r.estimated = datetime.timedelta(seconds=60)
        s = self.node.register_activity(r)
        self.assertTrue(s.starts is not None)
        s.unschedule()
        self.assertTrue(s.starts is None)

    def test_unschedule_ended_activity(self):
        r = Activity()
        r.estimated = datetime.timedelta(seconds=60)
        s = self.node.register_activity(r)
        self.assertTrue(s.starts is not None)
        s.end('OK', datetime.timedelta(seconds=70))
        self.assertRaises(RuntimeError, s.unschedule)
        self.assertEquals(
            datetime.datetime(2011, 7, 2, 22, 14, 0, 0, pytz.UTC), s.starts)
        self.assertEquals(True, s.ended)
        self.assertEquals('OK', s.result)
        self.assertEquals(datetime.timedelta(seconds=70), s.duration)

    def test_leadtime_requires_wrap_to_next_day(self):
        self.now.return_value = datetime.datetime(
            2011, 7, 8, 21, 0, 0, 0, pytz.UTC)
        self.node.resource_group.notification_leadtime = 8
        self.node.resource_group.maintenance_start = 22
        self.node.resource_group.maintenance_end = 5
        estimated_duration = datetime.timedelta(seconds=1 * 60)

        expected_range = [
            datetime.datetime(2011, 7, 9, 22, 0, 0, 0, pytz.UTC),
            datetime.datetime(2011, 7, 10, 4, 59, 0, 0, pytz.UTC)]
        request = Activity()
        request.estimated = estimated_duration
        suggestion = self.node.register_activity(request).starts
        self.assertInRange(expected_range, suggestion)

    def test_multi_day_leadtime(self):
        self.now.return_value = datetime.datetime(
            2012, 8, 6, 21, 0, 0, 0, pytz.UTC)
        self.node.resource_group.notification_leadtime = 96
        self.node.resource_group.maintenance_start = 22
        self.node.resource_group.maintenance_end = 5
        estimated_duration = datetime.timedelta(seconds=1 * 60)

        expected_range = [
            datetime.datetime(2012, 8, 10, 22, 0, 0, 0, pytz.UTC),
            datetime.datetime(2012, 8, 11, 4, 59, 0, 0, pytz.UTC)]
        request = Activity()
        request.estimated = estimated_duration
        suggestion = self.node.register_activity(request).starts
        self.assertInRange(expected_range, suggestion)

    def test_estimated_longer_than_allowed_gives_no_suggestion(self):
        # Allow 1 hour daily maintenance
        self.node.resource_group.maintenance_start = 22
        self.node.resource_group.maintenance_end = 23
        # Requesting 24 hours of maintenance results in no suggestion.
        request = Activity()
        request.estimated = datetime.timedelta(days=1)
        suggestion = self.node.register_activity(request).starts
        self.assertEquals(None, suggestion)

    def test_estimated_exactly_uses_full_window(self):
        # Allow 1 hour daily maintenance
        self.node.resource_group.maintenance_start = 22
        self.node.resource_group.maintenance_end = 23
        # Requesting 1 hour of maintenance results in first hit.
        request = Activity()
        request.estimated = datetime.timedelta(seconds=60 * 60)
        suggestion = self.node.register_activity(request).starts
        self.assertEquals(datetime.datetime(2011, 7, 2, 22, 0, 0, 0, pytz.UTC),
                          suggestion)

    def test_rescheduling_without_change_keeps_assignment_stable(self):
        # This is a test for a regression we had: once an activity is
        # scheduled it may not be updated to an estimated size that allows it
        # to overflow the allocated window size but must return with `None`
        # instead.
        self.now.return_value = datetime.datetime(
            2011, 7, 8, 10, 0, 0, 0, pytz.UTC)
        self.node.resource_group.notification_leadtime = 4
        self.node.resource_group.maintenance_start = 21
        self.node.resource_group.maintenance_end = 23

        request1 = Activity()
        request1.estimated = datetime.timedelta(seconds=30 * 60)
        suggestion1 = self.node.register_activity(request1)
        self.assertEquals(
            datetime.datetime(2011, 7, 8, 21, 14, tzinfo=pytz.UTC),
            suggestion1.starts)
        request1a = Activity()
        request1a.estimated = datetime.timedelta(seconds=30 * 60)
        suggestion1a = self.node.register_activity(request1a)
        self.assertEquals(
            datetime.datetime(2011, 7, 8, 21, 14, tzinfo=pytz.UTC),
            suggestion1.starts)
        self.assertEquals(
            suggestion1a.starts, suggestion1.starts)

    def test_rescheduling_does_not_allow_overflow(self):
        # This is a test for a regression we had: once an activity is
        # scheduled it may not be updated to an estimated size that allows it
        # to overflow the allocated window size but must return with `None`
        # instead.
        self.now.return_value = datetime.datetime(
            2011, 7, 8, 10, 0, 0, 0, pytz.UTC)
        self.node.resource_group.notification_leadtime = 4
        self.node.resource_group.maintenance_start = 21
        self.node.resource_group.maintenance_end = 23

        request1 = Activity()
        request1.estimated = datetime.timedelta(seconds=30 * 60)
        suggestion1 = self.node.register_activity(request1)
        self.assertEquals(
            datetime.datetime(2011, 7, 8, 21, 14, tzinfo=pytz.UTC),
            suggestion1.starts)
        request1a = Activity()
        request1a.estimated = datetime.timedelta(seconds=4 * 60 * 60)
        suggestion1a = self.node.register_activity(request1a)
        self.assertEquals(None, suggestion1a.starts)
        # As suggestion1a and suggestion1 should be the same objects
        # now suggestion1's start should have been reset, too.
        self.assertEquals(None, suggestion1.starts)

    def test_window_sorted_activities(self):
        request1 = Activity()
        request1.estimated = datetime.timedelta(seconds=30 * 60)
        suggestion1 = self.node.register_activity(request1)
        self.assertEquals([suggestion1], suggestion1.window.sorted_activities)

    def test_cancel_window_unschedules_activities(self):
        request1 = Activity()
        request1.estimated = datetime.timedelta(seconds=30 * 60)
        suggestion1 = self.node.register_activity(request1)
        window = suggestion1.window
        self.assertTrue(suggestion1.starts is not None)
        self.assertTrue(window.node is not None)
        suggestion1.window.cancel()
        self.assertEquals(None, suggestion1.window)
        self.assertEquals(None, suggestion1.starts)
        self.assertEquals(None, window.node)

    def test_windows_with_ended_activities_cant_be_cancelled(self):
        request1 = Activity()
        request1.estimated = datetime.timedelta(seconds=30 * 60)
        request1.comment = 'test'
        request1.reference_id = 'A'
        suggestion1 = self.node.register_activity(request1)
        window = suggestion1.window
        self.assertTrue(window.can_be_cancelled)
        suggestion1.end('OK')
        self.assertFalse(window.can_be_cancelled)

    def test_window_and_activity_timezones_are_in_utc(self):
        self.node.resource_group.timezone = 'Europe/Berlin'
        berlin = pytz.timezone('Europe/Berlin')
        self.node.resource_group.maintenance_start = 6
        self.node.resource_group.maintenance_end = 8
        self.node.resource_group.notification_leadtime = 1
        # The current time on this day is well before the potential start
        # today's period so we will end up with the window on this day
        self.now.return_value = datetime.datetime(
            2011, 7, 1, 1, 0, 0, 0, pytz.UTC)
        request1 = Activity()
        request1.estimated = datetime.timedelta(seconds=30 * 60)
        suggestion1 = self.node.register_activity(request1)
        # The window in Europe/Berlin starts at 6, which is 4 UTC today:
        self.assertEquals(
            datetime.datetime(2011, 7, 1, 4, 0, 0, 0, pytz.UTC),
            suggestion1.window.starts)
        self.assertEquals(
            berlin.localize(datetime.datetime(2011, 7, 1, 6, 0, 0, 0)),
            suggestion1.window.starts)
        self.assertEquals(pytz.UTC, suggestion1.window.starts.tzinfo)

        # The window in Europe/Berlin ends at 8, which is 6 UTC today:
        self.assertEquals(
            datetime.datetime(2011, 7, 1, 6, 0, 0, 0, pytz.UTC),
            suggestion1.window.ends)
        self.assertEquals(
            berlin.localize(datetime.datetime(2011, 7, 1, 8, 0, 0, 0)),
            suggestion1.window.ends)
        self.assertEquals(pytz.UTC, suggestion1.window.ends.tzinfo)

        # The activity starts like the window at 6 Europe/Berlin and 4 UTC:
        self.assertEquals(
            datetime.datetime(2011, 7, 1, 4, 14, 0, 0, pytz.UTC),
            suggestion1.starts)
        self.assertEquals(
            berlin.localize(datetime.datetime(2011, 7, 1, 6, 14, 0, 0)),
            suggestion1.starts)
        self.assertEquals(pytz.UTC, suggestion1.starts.tzinfo)

        # The activity ends 30 minutes later 6:30 Europe/Berlin and 4:30 UTC:
        self.assertEquals(
            datetime.datetime(2011, 7, 1, 4, 44, 0, 0, pytz.UTC),
            suggestion1.ends)
        self.assertEquals(
            berlin.localize(datetime.datetime(2011, 7, 1, 6, 44, 0, 0)),
            suggestion1.ends)
        self.assertEquals(pytz.UTC, suggestion1.ends.tzinfo)

    def test_started_unannounced_windows_dont_allow_scheduling(self):
        self.now.return_value = datetime.datetime(
            2011, 7, 8, 10, 0, 0, 0, pytz.UTC)
        window = Window()
        window.starts = datetime.datetime(2011, 7, 8, 9, 0, 0, 0, pytz.UTC)
        window.ends = datetime.datetime(2011, 7, 8, 11, 0, 0, 0, pytz.UTC)
        window.node = self.node
        self.assertEquals(False, window._scheduling_allowed)

    def test_started_announced_windows_allow_scheduling(self):
        self.now.return_value = datetime.datetime(
            2011, 7, 8, 10, 0, 0, 0, pytz.UTC)
        window = Window()
        window.starts = datetime.datetime(2011, 7, 8, 9, 0, 0, 0, pytz.UTC)
        window.ends = datetime.datetime(2011, 7, 8, 11, 0, 0, 0, pytz.UTC)
        window.announced = True
        window.node = self.node
        self.assertEquals(True, window._scheduling_allowed)

    def test_windows_allow_scheduling_after_leadtime_if_announced(self):
        self.now.return_value = datetime.datetime(
            2011, 7, 8, 21, 0, 0, 0, pytz.UTC)
        window = Window()
        window.starts = datetime.datetime(2011, 7, 8, 22, 0, 0, 0, pytz.UTC)
        window.ends = datetime.datetime(2011, 7, 8, 22, 30, 0, 0, pytz.UTC)
        window.announced = True
        window.node = self.node
        self.node.resource_group.notification_leadtime = 2
        self.node.resource_group.maintenance_start = 22
        self.node.resource_group.maintenance_end = 23
        self.assertEquals(True, window._scheduling_allowed)

    def test_schedule_activity_after_other_even_if_not_overlapping(self):
        self.now.return_value = datetime.datetime(
            2011, 7, 8, 10, 0, 0, 0, pytz.UTC)
        window = Window()
        window.starts = datetime.datetime(2011, 7, 8, 23, 0, 0, 0, pytz.UTC)
        window.ends = datetime.datetime(2011, 7, 9, 0, 0, 0, 0, pytz.UTC)
        window.node = self.node
        activity = Activity()
        activity.estimated = datetime.timedelta(seconds=10 * 60)
        activity.node = self.node
        activity.starts = datetime.datetime(2011, 7, 8, 23, 30, 0, 0, pytz.UTC)
        window.activities.append(activity)
        activity2 = Activity()
        activity2.estimated = datetime.timedelta(seconds=10 * 60)
        activity2.node = self.node
        window.schedule(activity2)
        self.assertTrue(activity2.starts is not None)
        self.assertTrue(activity2.starts > activity.starts)

    def test_window_ended_if_in_past_and_all_activities_ended(self):
        window = Window()
        window.starts = datetime.datetime(2011, 7, 8, 23, 0, 0, 0, pytz.UTC)
        window.ends = datetime.datetime(2011, 7, 9, 0, 0, 0, 0, pytz.UTC)
        window.node = self.node
        activity1 = Activity()
        activity1.estimated = datetime.timedelta(seconds=10 * 60)
        activity1.node = self.node
        window.schedule(activity1)
        activity2 = Activity()
        activity2.estimated = datetime.timedelta(seconds=10 * 60)
        activity2.node = self.node
        window.schedule(activity2)
        activity1.end('OK')
        activity2.end('OK')
        self.now.return_value = datetime.datetime(
            2011, 7, 9, 10, 0, 0, 0, pytz.UTC)
        self.assertTrue(window.ended)

    def test_window_not_ended_if_end_in_future(self):
        self.now.return_value = datetime.datetime(
            2011, 7, 8, 10, 0, 0, 0, pytz.UTC)
        window = Window()
        window.starts = datetime.datetime(2011, 7, 8, 23, 0, 0, 0, pytz.UTC)
        window.ends = datetime.datetime(2011, 7, 9, 0, 0, 0, 0, pytz.UTC)
        self.assertFalse(window.ended)

    def test_window_not_ended_if_end_in_past_but_activity_not_ended(self):
        self.now.return_value = datetime.datetime(
            2011, 7, 8, 10, 0, 0, 0, pytz.UTC)
        window = Window()
        window.starts = datetime.datetime(2011, 7, 8, 23, 0, 0, 0, pytz.UTC)
        window.ends = datetime.datetime(2011, 7, 9, 0, 0, 0, 0, pytz.UTC)
        window.node = self.node
        activity1 = Activity()
        activity1.estimated = datetime.timedelta(seconds=10 * 60)
        activity1.node = self.node
        window.schedule(activity1)
        self.now.return_value = datetime.datetime(
            2011, 7, 9, 10, 0, 0, 0, pytz.UTC)
        self.assertFalse(window.ended)

    def test_activity_local_starts_correct_timezone(self):
        activity = Activity()
        activity.estimated = datetime.timedelta(seconds=10 * 60)
        self.node.register_activity(activity)
        self.node.resource_group.timezone = 'Europe/Berlin'
        self.assertEquals(activity.starts, activity.starts_local)
        self.assertEquals(pytz.timezone('Europe/Berlin').zone,
                          activity.starts_local.tzinfo.zone)

    def test_activity_local_starts_handles_none(self):
        activity = Activity()
        activity.estimated = datetime.timedelta(seconds=10 * 60)
        self.assertEquals(None, activity.starts)
        self.assertEquals(None, activity.starts_local)

    def test_planning_unfinished_window_listing_no_windows(self):
        self.assertEquals([], self.node.unfinished_windows)

    def test_planning_unfinished_window_listing(self):
        window = Window()
        window.starts = datetime.datetime(2011, 7, 8, 23, 0, 0, 0, pytz.UTC)
        window.ends = datetime.datetime(2011, 7, 9, 0, 0, 0, 0, pytz.UTC)
        window._materialize(self.node)
        self.assertEquals([window], self.node.unfinished_windows)

    def test_planning_unfinished_window_listing_finished_windows(self):
        self.now.return_value = datetime.datetime(
            2011, 7, 9, 10, 0, 0, 0, pytz.UTC)
        window = Window.create()
        window.starts = datetime.datetime(2011, 7, 8, 23, 0, 0, 0, pytz.UTC)
        window.ends = datetime.datetime(2011, 7, 9, 0, 0, 0, 0, pytz.UTC)
        window._materialize(self.node)
        self.assertEquals([], self.node.unfinished_windows)

    def test_planning_list_ended_activities_no_activitites(self):
        self.assertEquals([], self.node.ended_activities)

    def test_planning_list_ended_activities_open_activities(self):
        activity = Activity()
        activity.estimated = datetime.timedelta(seconds=10 * 60)
        self.node.register_activity(activity)
        self.assertFalse(activity.ended)
        self.assertEquals([], self.node.ended_activities)

    def test_planning_list_ended_activities(self):
        activity = Activity()
        activity.estimated = datetime.timedelta(seconds=10 * 60)
        self.node.register_activity(activity)
        activity.end('OK')
        self.assertEquals([activity], self.node.ended_activities)

    def test_schedule_activity_in_europe_oslo_correct_offset_during_dst(self):
        # This is a regression against #9893
        self.now.return_value = datetime.datetime(
            2011, 11, 3, 10, 0, 0, 0, pytz.UTC)
        self.node.resource_group.notification_leadtime = 4
        self.node.resource_group.maintenance_start = 22
        self.node.resource_group.maintenance_end = 4
        self.node.resource_group.timezone = 'Europe/Oslo'
        estimated_duration = datetime.timedelta(seconds=30 * 60)

        r = Activity()
        r.estimated = estimated_duration
        s = self.node.register_activity(r)
        self.assertEquals(
            datetime.datetime(2011, 11, 3, 21, 14, 0, 0, pytz.UTC),
            s.starts)

    def test_announcement_should_contain_comment_and_servicedesc(self):
        w = Window()
        w.starts = self.now()
        w.ends = self.now()
        w.comment = 'Window for testing purposes only.'
        self.node.register_window(w)
        n = gocept.directory2.maintenance.Notifier()
        subj, body = n.render_message(w)
        self.assert_(w.comment in body, 'comment not in body\n%s' % body)
        self.assert_(self.node.service_description in body,
                     'service_description not in body\n' + body)

    def test_announcement_should_contain_activities(self):
        self.node.service_description = None
        request = Activity()
        request.estimated = datetime.timedelta(minutes=2)
        request.comment = 'test activity'
        self.node.register_activity(request)
        n = gocept.directory2.maintenance.Notifier()
        subj, body = n.render_message(request.window)
        self.assert_('test activity' in body,
                     'activity description not in body\n' + body)

    def test_optimize_followup_activities(self):
        # Assume it's 2011-07-08 11:30 now.
        self.now.return_value = datetime.datetime(
            2011, 7, 8, 11, 30, 0, 0, pytz.UTC)

        # Assume that we have a window scheduled between 11:00 and 18:00
        window = Window()
        window.starts = datetime.datetime(2011, 7, 8, 11, 0, 0, 0, pytz.UTC)
        window.ends = datetime.datetime(2011, 7, 9, 18, 0, 0, 0, pytz.UTC)
        window.node = self.node
        window.announced = True

        # Assume that we have a first activity schedule directly for 11:00
        # with a duration of 60 minutes
        activity = Activity()
        activity.estimated = datetime.timedelta(seconds=60 * 60)
        activity.node = self.node
        window.schedule(activity)

        # A second activity with 10 minutes gets scheduled and will get a start
        # date after 12:00
        activity2 = Activity()
        activity2.estimated = datetime.timedelta(seconds=10 * 60)
        activity2.node = self.node
        window.schedule(activity2)
        self.assertIsNotNone(activity2.starts)
        self.assertGreaterEqual(
            activity2.starts,
            datetime.datetime(2011, 7, 8, 12, 00, 0, 0, pytz.UTC))

        # A third activity with 20 minutes gets scheduled and will get a
        # start date after 12:10
        activity3 = Activity()
        activity3.estimated = datetime.timedelta(seconds=10 * 60)
        activity3.node = self.node
        window.schedule(activity3)
        self.assertIsNotNone(activity3.starts)
        self.assertGreaterEqual(
            activity3.starts,
            datetime.datetime(2011, 7, 8, 12, 10, 0, 0, pytz.UTC))
        # Now, it's 11:30 and we get feedback that the first activity has
        # finished already.
        activity.end('OK', datetime.timedelta(seconds=70))

        # This results in the second activity to be scheduled 'right away'
        # at the end of the first activity and the third at the projected
        # end of the second.
        self.assertEqual(activity2.starts,
                         activity.starts + activity.duration)
        self.assertEqual(activity3.starts, activity2.ends)


class MaintenanceWeekTests(unittest.TestCase):

    def setup(self, strategy_factory, node_count=1):
        self.s = strategy_factory()
        self.s.first_day = datetime.datetime(2013, 1, 21)
        self.s.nodes = [mock.Mock() for i in range(node_count)]
        self.node_count = node_count

        # This node is prepared for testing select()
        self.node = mock.Mock()
        self.node.puppet_classes = set([])

    def plan(self):
        self.events = list(self.s.plan())
        self.events.sort(key=lambda x: x.start)
        self.assertEqual(self.node_count, len(self.events))

    def test_restricted_strategy(self):
        self.setup(gocept.directory2.maintenance.Restricted)

        self.assertFalse(self.s.select(self.node))

        self.node.puppet_classes = ['role::restricted']
        self.assertTrue(self.s.select(self.node))

        self.plan()

        self.assertEquals(datetime.datetime(2013, 1, 24, 17),
                          self.events[0].start)

    def test_customer_physical(self):
        self.setup(gocept.directory2.maintenance.CustomerPhysical)

        self.node.directory_ring = 0
        self.assertFalse(self.s.select(self.node))

        self.node.directory_ring = 1
        self.assertTrue(self.s.select(self.node))

        self.plan()

        self.assertEquals(datetime.datetime(2013, 1, 24, 17),
                          self.events[0].start)

    def test_generic_devices(self):
        self.setup(gocept.directory2.maintenance.GenericDevices)

        self.assertTrue(self.s.select(self.node))

        self.node.puppet_classes = ['role::generic']
        self.assertTrue(self.s.select(self.node))

        self.node.puppet_classes = ['role::foobar']
        self.assertFalse(self.s.select(self.node))

        self.plan()
        self.assertEquals(datetime.datetime(2013, 1, 26, 17),
                          self.events[0].start)

    def test_storages(self):
        self.setup(gocept.directory2.maintenance.Storages, 10)

        self.assertFalse(self.s.select(self.node))

        self.node.puppet_classes = ['role::storageserver']
        self.assertTrue(self.s.select(self.node))

        self.plan()
        self.assertEquals(datetime.datetime(2013, 1, 21, 13),
                          self.events[0].start)
        self.assertEquals(datetime.datetime(2013, 1, 22, 17),
                          self.events[1].start)
        self.assertEquals(datetime.datetime(2013, 1, 23, 17),
                          self.events[2].start)
        self.assertEquals(datetime.datetime(2013, 1, 23, 17),
                          self.events[3].start)
        self.assertEquals(datetime.datetime(2013, 1, 24, 17),
                          self.events[4].start)
        self.assertEquals(datetime.datetime(2013, 1, 24, 17),
                          self.events[5].start)
        self.assertEquals(datetime.datetime(2013, 1, 24, 17),
                          self.events[6].start)
        self.assertEquals(datetime.datetime(2013, 1, 24, 17),
                          self.events[7].start)
        self.assertEquals(datetime.datetime(2013, 1, 24, 17),
                          self.events[8].start)
        self.assertEquals(datetime.datetime(2013, 1, 24, 17),
                          self.events[9].start)

    def test_storages_only_1(self):
        self.setup(gocept.directory2.maintenance.Storages, 1)
        self.plan()

    def test_storages_only_2(self):
        self.setup(gocept.directory2.maintenance.Storages, 2)
        self.plan()

    def test_storages_only_3(self):
        self.setup(gocept.directory2.maintenance.Storages, 3)
        self.plan()

    def test_routers(self):
        self.setup(gocept.directory2.maintenance.Routers, 2)

        self.assertFalse(self.s.select(self.node))

        self.node.puppet_classes = ['role::router']
        self.assertTrue(self.s.select(self.node))

        self.plan()
        self.assertEquals(datetime.datetime(2013, 1, 21, 13),
                          self.events[0].start)
        self.assertEquals(datetime.datetime(2013, 1, 26, 13),
                          self.events[1].start)

    def test_kvm_hosts(self):
        self.setup(gocept.directory2.maintenance.KVMHosts, 2)

        self.assertFalse(self.s.select(self.node))

        self.node.puppet_classes = ['role::kvm_host']
        self.assertTrue(self.s.select(self.node))

        self.plan()
        self.assertEquals(datetime.datetime(2013, 1, 26, 9),
                          self.events[0].start)
        self.assertEquals(datetime.datetime(2013, 1, 26, 9),
                          self.events[1].start)

    def test_production_vms(self):
        self.setup(gocept.directory2.maintenance.ProductionVMs, 20)

        self.assertFalse(self.s.select(self.node))

        vm = VirtualMachine()
        self.assertTrue(self.s.select(vm))

        self.s.nodes[0].location.directory.resource_groups = []

        self.plan()

        # 5% few on tuesday = 1
        self.assertEquals(datetime.datetime(2013, 1, 22, 17),
                          self.events[0].start)

        # 30% more on wednesday = 6
        for i in range(1, 7):
            self.assertEquals(datetime.datetime(2013, 1, 23, 17),
                              self.events[i].start)

        # All remaining on thursday
        for i in range(7, 20):
            self.assertEquals(datetime.datetime(2013, 1, 24, 17),
                              self.events[i].start)

    def test_customer_test_machines(self):
        self.setup(gocept.directory2.maintenance.CustomerTestMachines, 10)

        self.node.name = 'oshaweb00'
        self.assertFalse(self.s.select(self.node))

        for name in ['oshawebtest00', 'oshawebdev00',
                     'oshawebstag00', 'oshawebstaging00']:
            self.node.name = name
            self.assertTrue(self.s.select(self.node))

        self.s.nodes[0].puppet_classes = set(['database'])
        self.s.nodes[1].puppet_classes = set(['application1', 'application2'])
        self.s.nodes[2].puppet_classes = set(['database2', 'application3'])
        self.s.nodes[3].puppet_classes = set(['frontend', 'application4'])
        for i in range(4, 10):
            self.s.nodes[i].puppet_classes = set(['application1'])

        self.plan()

        # Representative set is due on Monday
        for i in range(4):
            self.assertEquals(datetime.datetime(2013, 1, 21, 17),
                              self.events[i].start)

        # All other test VMs on Tuesday
        for i in range(4, 10):
            self.assertEquals(datetime.datetime(2013, 1, 22, 17),
                              self.events[i].start)

    def test_internal_test_machines(self):
        self.setup(gocept.directory2.maintenance.InternalTestMachines, 2)

        self.node.name = 'oshaweb00'
        self.assertFalse(self.s.select(self.node))

        for name in ['oshawebtest00', 'oshawebdev00',
                     'oshawebstag00']:
            self.node.name = name
            self.assertFalse(self.s.select(self.node))

        for name in ['test00', 'litdev00', 'litprod00', 'litstag00']:
            self.node.name = name
            self.assertTrue(self.s.select(self.node))

        self.plan()

        self.assertEquals(datetime.datetime(2013, 1, 21, 17),
                          self.events[0].start)
        self.assertEquals(datetime.datetime(2013, 1, 21, 17),
                          self.events[1].start)

    def test_early_infrastructure(self):
        self.setup(gocept.directory2.maintenance.EarlyInfrastructure, 2)

        self.assertFalse(self.s.select(self.node))

        roles = (gocept.directory2.maintenance.EarlyInfrastructure.
                 automation_classes)
        for role in roles:
            self.node.puppet_classes = set([role])
            self.assertTrue(self.s.select(self.node))

        self.plan()

        self.assertEquals(datetime.datetime(2013, 1, 21, 9),
                          self.events[0].start)
        self.assertEquals(datetime.datetime(2013, 1, 21, 9),
                          self.events[1].start)
